/* 
 * Copyright(c) 2013 Intel Corporation. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 * * Neither the name of Intel Corporation nor the names of its
 * contributors may be used to endorse or promote products derived
 * from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*!	\file CreateCSSInput.c
 * 	\brief	Tools provides an output file with the correct layout.
 * 	The output of this tools is used as an input into the next stage
 * 	which generates the hashing and cryptographic data.	
 */


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#ifdef _MSC_VER
typedef unsigned long uint32_t;
#else
#include <stdint.h>
#endif

#define RSA_MODULUS_SIZE_BYTES		256
#define RSA_SIGNATURE_SIZE_BYTES	256
#define BODY_OFFSET_MAX			0xFFFFFFFF // TODO: Decide on a maximum body offset
#define BODY_OFFSET_MIN 		0x24C


typedef struct ClantonSecurityHeader
{
	uint32_t CSH_Identifier;					// Unique identifier to sanity check signed module presence
    uint32_t Version;							// Currently unused for Clanton. Maintaining for future use
    uint32_t ModuleSizeBytes;					// Size of the entire module including the Clanton Security Header, RSA Key, RSA Signature and Body
    uint32_t SecurityVersionNumberIndex;		// Index of the SVN to be used for module validation
    uint32_t SecurityVersionNumber;				// Used to prevent against roll back of modules
    uint32_t RsvdModuleID;						// Currently unused for Clanton. Maintaining for future use
    uint32_t RsvdModuleVendor;					// Vendor Identifier. For Intel products value is 0x00008086
    uint32_t RsvdDate;							// BCD representation of build date as yyyymmdd, where yyyy=4 digit year, mm=1-12, dd=1-31
    uint32_t ModuleHeaderSizeBytes;				// Total length of the header including signed portion, RSA Key, Signature andany padding
    uint32_t HashAlgorithm;						// Hashing Algorithm used during the signing process
    uint32_t CryptoAlgorithm;					// Signing Algorithm used during signing process
    uint32_t KeySizeBytes;       				// Total length of the key data including and padding optionally added by signing tool
    uint32_t SignatureSizeBytes;        		// total length of the signature including any padding optionally added by signing tool
    uint32_t RsvdNextHeaderPointer;
    uint32_t Reserved[8/sizeof(uint32_t)];		// Bytes reserved for future use.
} ClantonSecurityHeader;


typedef struct RSAPublicKey
{
	uint32_t ModulusSizeBytes;
	uint32_t ExponentSizeBytes;
    uint32_t modulus[RSA_MODULUS_SIZE_BYTES/4];					// TODO: Resolve Hardcoded value here!
    uint32_t exponent;
} RSAPublicKey;

typedef struct CSSSigningFormat
{
    ClantonSecurityHeader Csh;
    RSAPublicKey SignatureVerificationKey;
    uint32_t Signature[RSA_SIGNATURE_SIZE_BYTES/4];
    char* Body;
} CSSSigningFormat;




#define    PERFORM_32BIT_ENDIAN_SWAP(input)\
        (((input>>24)&0xff) | ((input&0xff)<<24) | ((input&0xFF0000)>>8) | ((input<<8)&0xFF0000))

uint32_t ReadBinaryFile(const char* filename, char** output_buffer, int* output_buffer_size_bytes);
void sec_fw_memcpy(uint32_t* dest, uint32_t dest_byte_offset,
                   uint32_t* source, uint32_t source_byte_offset,
                   uint32_t num_bytes, uint32_t dword_endian_swap,
                   uint32_t reverse_memory_order);
uint32_t bcd_date(void);
void PrintData(char* string, void* dataInput, uint32_t num_bytes);
uint32_t bcd_date(void);
void usage();

int main(int argc, char *argv[])
{
    uint32_t bodyOffset = 0;
    uint32_t inputInHex = 0;
    uint32_t outputInHex = 0;
    uint32_t SvnIndex = 0;
    uint32_t Svn = 0;

    char* inputFilename = 0x0;
    char* outputFilename = 0x0;
    char* inputData = NULL;
    char* padding = NULL;
    int inputDataSizeBytes = 0;
    uint32_t badarg = 0;
    int num_args= argc;
    int errorCount = 0;
    FILE* outputFile = NULL;
    argc--;
    argv++;

    printf("Collecting Arguments\n");
    while(argc >= 1)
    {
        printf("Checking argument %d = %s\n", num_args - argc, *argv);
        if (!strcmp(*argv,"-input"))
        {
            argc -= 1;
            if (argc < 1)
                badarg = 1;
            else
            {
                argv++;		//now at filename
                badarg = 0;
                printf("Collecting input filename %s\n", *(argv));
                inputFilename = *(argv);
                
            }
        }
        else if(!strcmp(*argv, "-bodyoffset"))
        {
            argc -= 1;
            if (argc < 1)
                badarg = 1;
            else
            {
                printf("Collecting body offset %s\n", *(argv+1));
                sscanf(*(++argv), "%x", &bodyOffset);
		
		if((bodyOffset > BODY_OFFSET_MAX) && (bodyOffset < BODY_OFFSET_MIN))
		{
			fprintf(stderr,"Bad body offset: %x \n",bodyOffset);
			badarg = 1;
		}
				}
#if 0	// Removed to suppress klocwork warning. 
                if( bodyOffset != 0x24C)
                {
                	printf("Body Offset not 0x24C\n");
                	padding = (char*) calloc ( (bodyOffset-0x24C),sizeof(char));
                }
#endif
        }
        else if (!strcmp(*argv,"-output"))
        {
            argc -= 1;
            if (argc < 1)
            {
                badarg = 1;
            }
            else
            {
                badarg = 0;
                printf("Request for %s output file\n", *(++argv));
                outputFilename = *(argv);
            }
        }
        else if (!strcmp(*argv,"-svnindex"))
        {
            argc -= 1;
            if (argc < 1)
            {
                badarg = 1;
            }
            else
            {
                printf("SVN Index %s\n", *(argv+1));
                sscanf(*(++argv), "%x", &SvnIndex);
                
            }
        }
        else if (!strcmp(*argv,"-svn"))
        {
            argc -= 1;
            if (argc < 1)
            {
                badarg = 1;
            }
            else
            {
                printf("SVN  %s\n", *(argv+1));
                sscanf(*(++argv), "%x", &Svn);                
            }
        }
        else
            badarg = 1;

        if(badarg)
        {
            break;
        }

        argc--;
        argv++;
    }
    printf("Done Collecting Arguments\n");
    if (!badarg)
    {
        // Read the input file
        if (inputFilename != NULL)
        {
                if (!ReadBinaryFile(inputFilename, &inputData, &inputDataSizeBytes))
                {
                    printf("Cannot open input file\n");
                    badarg = 1;
                }
        }
        else
        {
            printf("Invalid Input filename\n");
            badarg = 1;
        }
        if (outputFilename != NULL)
        {
            if (outputInHex)
            {
                outputFile = fopen(outputFilename, "w");
            }
            else
            {
                outputFile = fopen(outputFilename, "wb");
            }
            if (outputFile == NULL)
            {
                printf("Cannot open output file\n");
                badarg = 1;
            }
        }
        else
        {
            printf("Invalid Output filename\n");
            badarg = 1;
        }
        if (bodyOffset == 0x0)
        {
            printf("Invalid Body Location\n");
            badarg = 1;
        }
    }

    if (!badarg)
    {
//        if (!(inputDataSizeBytes & (64-1)))
//        {
            CSSSigningFormat cssStructure;
            memset((char*)&cssStructure, 0x0, sizeof(cssStructure));
			cssStructure.Csh.CSH_Identifier = 0x5F435348;
            cssStructure.Csh.Version = 0x00000001;
            cssStructure.Csh.SecurityVersionNumberIndex = SvnIndex;
            cssStructure.Csh.SecurityVersionNumber = Svn;
            //cssStructure.Csh.ModuleHeaderSizeBytes = sizeof(RSAPublicKey) + sizeof(ClantonSecurityHeader) + sizeof(cssStructure.Signature);		// TODO: Fix Calculation here
            cssStructure.Csh.ModuleHeaderSizeBytes = bodyOffset;
            cssStructure.Csh.RsvdModuleID = 0x00000000;
			cssStructure.Csh.RsvdModuleVendor = 0x00008086;
            cssStructure.Csh.RsvdDate = 0x00000000;
            // This will change in Extended Range Authenticated Boot Mode
            cssStructure.Csh.ModuleSizeBytes = (inputDataSizeBytes + bodyOffset);
            cssStructure.Csh.HashAlgorithm = 0x00000001;
            cssStructure.Csh.CryptoAlgorithm = 0x00000001;
            cssStructure.Csh.KeySizeBytes = 268;		// TODO: Check that this hardcoded value is OK
            // This will change in Extended Range Authenticated Boot Mode
            cssStructure.Csh.SignatureSizeBytes = 256;		// TODO: Hardcoded at the moment
            cssStructure.Csh.RsvdNextHeaderPointer = 0x00000000;
            // This will change in 32KB Internal SRAM encrypted Boot Mode
			
            // Now write the output file
            cssStructure.Body = inputData;
            if( cssStructure.Body )
            {
                    printf("\nWriting binary output file %s\n", outputFilename);
                    printf("Header Size = %u bytes\n", (uint32_t)(cssStructure.Csh.ModuleHeaderSizeBytes));
                    fwrite((char*)&(cssStructure.Csh),sizeof(char), sizeof(cssStructure.Csh), outputFile);
                    fwrite((char*)&(cssStructure.SignatureVerificationKey), sizeof(char), sizeof(cssStructure.SignatureVerificationKey),outputFile);
                    fwrite((char*)&(cssStructure.Signature), sizeof(char), sizeof(cssStructure.Signature), outputFile);
                    if( bodyOffset != 0x24C)
                    {
                    	printf("Body Offset not 0x24C\n");

			if((bodyOffset <= BODY_OFFSET_MAX) && (bodyOffset > BODY_OFFSET_MIN)) // To suppress Klocwork warning - this value is also  checked when the arguments are collected.
			{
				padding = (char*) calloc ( (bodyOffset-0x24C),sizeof(char));
		
				if(padding != NULL)
				{		
					fwrite(padding, sizeof(char), bodyOffset-0x24C, outputFile);
				}
				else
				{
					fprintf(stderr,"ERROR: Cannot allocate memory for padding \n");
					errorCount++;
				}
			}
			else
			{	
				fprintf(stderr,"ERROR: Bad Body Offset \n");
				errorCount++;			}										
                    }
                    
                    
                    printf("Header writing complete\n");
                    printf("Body Size = %u bytes\n", (uint32_t)inputDataSizeBytes);
                    fwrite(cssStructure.Body, sizeof(char), inputDataSizeBytes, outputFile);
                    printf("Body writing complete\n");
            }
//        }
//        else
//        {
//            if (inputDataSizeBytes & (64-1))
//                printf("CSSInput:  ERROR: Input Data must be a multiple of 64 bytes.  Currently the filesize is 0x%08x bytes\n", inputDataSizeBytes);
//        }

        if (inputData)
            free(inputData);
	if (padding)
	    free(padding);
        fclose(outputFile);
	
	if (errorCount != 0)
	{
		//TODO: Remove output file if there was an error.
		fprintf(stderr,"ERROR! creating output file \n");
		return 1; 
	}
	else
	{
        	return 0;
	}

    }
    else
    {
        usage();

        if (inputData)
            free(inputData);
        if(outputFile)
            fclose(outputFile);

        return 1;
    }
}


uint32_t ReadBinaryFile(const char* filename, char** output_buffer, int* output_buffer_size_bytes)
{
    uint32_t fileGood = 1;
    int actualFileSize = 0x0;
    FILE* inputFile    = fopen(filename, "rb");
    if (inputFile != NULL)
    {
        printf("Scanning input file %s\n", filename);
        fseek(inputFile, 0, SEEK_END);
        actualFileSize = ftell(inputFile);
        fseek(inputFile, 0, SEEK_SET);

        printf("Binary filesize = %d bytes\n", actualFileSize);
        if (actualFileSize > 0)
        {
            *output_buffer = (char*)malloc(actualFileSize*sizeof(unsigned char));
            if(*output_buffer)
            {
                fread(*output_buffer, actualFileSize, sizeof(unsigned char), inputFile);
            }
        }
        fclose(inputFile);
    }
    else
        fileGood = 0;
    *output_buffer_size_bytes = actualFileSize;
    return fileGood;
}

void PrintData(char* string, void* dataInput, uint32_t num_bytes)
{
    uint32_t i;
    unsigned char* dataInputBytePointer = (unsigned char*)dataInput;
    printf("%s:",string);
    for (i=0; i<num_bytes; i++)
    {
        printf("%02x",(uint32_t)dataInputBytePointer[i]);
    }
    printf("\n");
}

void sec_fw_memcpy(uint32_t* dest, uint32_t dest_byte_offset,
                   uint32_t* source, uint32_t source_byte_offset,
                   uint32_t num_bytes, uint32_t dword_endian_swap,
                   uint32_t reverse_memory_order)
{
    uint32_t i;
    if ((dest_byte_offset == 0x0) && (source_byte_offset == 0x0) && ((num_bytes & 0x3) == 0x0))
    {
        // Efficient 32-bit memcpy
        for (i=0; i<num_bytes/4; i++)
        {
            uint32_t dest_index;
            uint32_t temp_data = source[i];
            temp_data = ((dword_endian_swap == 0x0) ? temp_data : PERFORM_32BIT_ENDIAN_SWAP(temp_data));
            dest_index = ((reverse_memory_order == 0x0) ? i : (num_bytes/4 - i)-1);
            dest[dest_index] = temp_data;
        }
    }
    else
    {
        // Byte wise memcpy
        for (i=0; i<num_bytes; i++)
        {
            uint32_t dest_index = (dest_byte_offset + i) / 4;
            uint32_t source_index = (source_byte_offset + i) / 4;
            uint32_t temp_byte = source[source_index];
            temp_byte = (temp_byte >> ((3-(source_byte_offset + i))%4)*8) & 0xFF;
            dest[dest_index] &= ~(0xFF000000 >> ((dest_byte_offset + i)%4)*8);
            dest[dest_index] |= temp_byte << ((3-(dest_byte_offset + i))%4)*8;
        }
    }
}

uint32_t bcd_date(void)
{
    struct tm *current_time;
    time_t t;
    uint32_t result = 0;
    int val;

    time(&t);
    current_time = localtime(&t);
    
    if(current_time)
    {
        val = current_time->tm_year + 1900;
        result |= (val/1000) << 28;
        val %= 1000;
        result |= (val/100) << 24;
        val %= 100;
        result |= (val/10) << 20;
        val %= 10;
        result |= val << 16;

        val = current_time->tm_mon + 1;
        result |= (val/10) << 12;
        val %= 10;
        result |= val << 8;

        val = current_time->tm_mday;
        result |= (val/10) << 4;
        val %= 10;
        result |= val;
    }

    return result;
}

void usage()
{
    fprintf(stderr, "Usage: cssinput [options]\n");
    fprintf(stderr, "-input [asciihex|bin] [be|le] file \n\t input (body) file [32-bit BigEndian | 32-bit LittleEndian]\n");
	fprintf(stderr, "-bodyoffset hexaddress \n\t hex offset of input file\n");
	fprintf(stderr, "-svnindex hexvalue \n\t [OPTIONAL]hex index of SVN to use for module validation, zero if not specified\n");
	fprintf(stderr, "-svn hexvalue \n\t [OPTIONAL]hex value of SVN to use for module validation, zero if not specified");
	fprintf(stderr, "-output file \n\t output file\n");


}
